#!/usr/bin/perl

######################################################################
package Cognition::HTMLParser;
######################################################################

use utf8;
use Cognition::Export::Entity;
use Cognition::HTMLParser::DocModel;
use Cognition::HTMLParser::eRDF;
use Cognition::HTMLParser::RDFa;
use Cognition::HTMLParser::RDF;
use Cognition::HTMLParser::RoleAttr;
use Cognition::HTMLParser::Metadata;
use Cognition::HTMLParser::Utils;
use Cognition::MagicString;
use Cognition::RDFModel::Property;
use Cognition::RDFModel::Subject;
use Cognition::RDFModel::Value;
use Cognition::uF::COinS;
use Cognition::uF::RelEnclosure;
use Cognition::uF::RelLicense;
use Cognition::uF::RelTag;
use Cognition::uF::XFN;
use Cognition::uF::adr;
use Cognition::uF::figure;
use Cognition::uF::geo;
use Cognition::uF::hAlarm;
use Cognition::uF::hAtom;
use Cognition::uF::hAudio;
use Cognition::uF::hCalendar;
use Cognition::uF::hCard;
use Cognition::uF::hEvent;
use Cognition::uF::hMeasure;
use Cognition::uF::hReview;
use Cognition::uF::hResume;
use Cognition::uF::hTodo;
use Cognition::uF::species;
use Cognition::uF::xoxo;
use Encode;
use HTML::Entities qw(:DEFAULT encode_entities_numeric %entity2char);
use HTML::Tagset;
use HTML::TreeBuilder;
use RDF::Redland;
use URI::URL;
use XML::LibXML qw(:all);
use strict;

BEGIN
{
	our $VERSION = 0.1;
	our $XDP;
}

sub new
{
	my $class    = shift;
	my $source   = shift;
	my $opts     = shift;

	my $this   = {
		'Source'   => $source,
		'UA'       => $opts->{ua},
		'REQUEST'  => $opts->{request},
		'RESPONSE' => $opts->{response},
		'options'  => {
			'erdf_strict_profiles'  => 1,
			'grddl_fetch'  => 0,
			'grddl_strict_profiles' => 0,
			'p_AUTO'       => 1,
			'p_erdf'       => 1,
			'p_grddl'      => 1,
			'p_rdfx'       => 1,
			'p_rdf'        => 1,
			'p_comments'   => 1,
			'p_rdfa'       => 1,
			'p_structure'  => 1,
			'p_uf'         => 1,
			'rdfa_strict_doctype'   => 0,
			'rdfa_strict_version'   => 0,
			'uf_strict_profiles'    => 0
		},
	};

	bless $this, $class;

	# Process options.
	foreach my $o (keys %$opts)
	{
		$this->{options}->{$o} = $opts->{$o}
			if (defined $opts->{$o} && defined $this->{options}->{$o});
	}
	
	foreach my $h (@{ $opts->{nofollow} })
	{
		$this->{options}->{nofollow}->{$h} = 1;
	}

	# Saved configuration stuff.
	$this->load_stored_data('subPropertyOf');

	# Infrastructure stuff
	$this->build_dom_tree;
	$this->find_base_uri;
	Cognition::HTMLParser::Utils::PrepareXPointers($this->{DOM});
	Cognition::HTMLParser::Utils::PrepareXmlLang($this->{DOM}->documentElement, undef, $opts->{response});
	$this->find_namespaces;
	$this->find_page_profiles;

	# Parse metadata automatically if that option is enabled.
	if ($this->{options}->{p_AUTO})
	{
		$this->parse_metadata;
		$this->parse_uf;
	}

	return $this;	
}


sub _fq2pfx
{
	my $this = shift;
	my $key  = shift;
	my $mode = shift || ':';
	
	my ($nsurl, $preferred_pfx);
	$key =~ s/(^\<|\>$)//g;
	if ($key =~ /^(.*[\#\/])([^\#\/]+)$/)
	{
		$nsurl = $1;
		$key   = $2;
	}
	else
	{
		return;
	}

	if (defined $this->{NamespacePrefixes}->{$nsurl})
		{ return $this->{NamespacePrefixes}->{$nsurl}.$mode.$key; }

	foreach my $ns (@{ $this->{Namespaces} })
	{
		$preferred_pfx = $ns->{prefix}
			if ($ns->{uri} eq $nsurl);
	}

	if ($preferred_pfx)
	{
		my $pfx = $preferred_pfx;
		while (defined $this->{NamespacePrefixesReverse}->{$pfx})
			{ $pfx = $preferred_pfx . int(rand(999)); }

		$this->{NamespacePrefixesReverse}->{$pfx} = $nsurl;
		$this->{NamespacePrefixes}->{$nsurl}      = $pfx;

		return $pfx.$mode.$key;
	}
	
	return $key;
}


sub _pfx2fq
{
	my $this = shift;
	my $key  = shift;
	my $mode = shift;
	my $dom  = shift;
	my ($ns, $term, $nsurl);
	
	$mode = '.:' unless length $mode;
	
	if (defined $dom && ($mode eq '.:') && length ($dom->getAttribute('rel').$dom->getAttribute('rev').$dom->getAttribute('property').$dom->getAttribute('datatype')))
	{
		# From HTML 4.01 and (in all-caps) HTML 3.2 specs 
		if ($key =~ /^\:?( Alternate | Stylesheet | Start | Next | Prev |
			Contents | Index | Glossary | Copyright | Chapter | Section |
			Subsection | Appendix | Help | Bookmark | PREVIOUS | SEARCH |
			MADE | TOP )$/ix)
		{
			$key = lc("xhv:$1");
		}
		# And the HTML 5 spec.
		# 2008-07-20 : removed 'contact' as it's gone from HTML 5 draft.
		elsif ($key =~ /^\:?( alternate | archives | author | bookmark |
			external | feed | first | help | icon | index | last | license | next |
			nofollow | noreferrer | pingback | prefetch | prev | search |
			stylesheet | sidebar | tag | up )$/ix)
		{
			$key = lc("xhv:$1");
		}
		# Selected values from XHTML 2
		elsif ($key =~ /^\:?( cite | meta | p3pv1 | role )$/ix)
		{
			$key = lc("xhv:$1");
		}
		# GRDDL
		elsif (lc($key) eq 'transformation')
		{
			return 'http://www.w3.org/2003/g/data-view#'.$key;
		}
		# Any other colon-prefixed names are put explicitly into the XHTML vocab.
		elsif ($key =~ /^\:/)
		{
			$key = "xhv$key";
		}
	}
	
	# In "dash mode" allows the hyphen to be used as a namespace
	# prefix separator. This is for eRDF.
	if (($mode eq '-' && $key =~ /^([A-Z_][A-Z0-9_]*)[\.\:-](.+)$/i)
	# Normally, dash is allowed as part of the namespace prefix.    
	|| ($mode eq '.:' && $key =~ /^([A-Z_][A-Z0-9_-]*)[\.\:](.+)$/i))
	{
		$ns = $1;
		$term = $2;
		my $xp = undef;
		$xp = $dom->getAttribute('_xpath') if ($dom);
		
		my $NS = Cognition::Namespace::bestMatch($ns, $xp, @{$this->{Namespaces}});
		$nsurl = $NS->{uri};
		
		# In eRDF, cope with undefined prefixes by assuming that
		# the class name was never intended as eRDF.
		if ($mode eq '-' && !$nsurl)
		{
			return $key;
		}
		
		# Otherwise, if an undefined prefix is used, complain quietly.
		$nsurl = 'http://undefined-namespace-prefix.invalid/' if (!$nsurl);
		
		return $nsurl.$term;
	}

	# If no valid prefix is found, treat as case-insensitive default namespace
	return $key;
}


sub abs_url
{
	my $this  = shift;
	my $param = shift || '';
	my $dom   = shift;
	
	# Support for CURIEs.
	if ($param =~ /^\[([A-Za-z0-9_]+)\:([A-Za-z0-9_]+)\]$/)
	{	
		my $ns    = $1;
		my $curie = $2;
		
		if ($ns eq '_')
			{ return 'bnode:///'.$curie; }

		my $expanded = $this->_pfx2fq("$ns:$curie", undef, $dom);
		return $expanded if (length $expanded);
	}
	elsif ($param eq '_:' || $param eq '[_:]')
	{
		return 'bnode:///iHateThisSyntax';
	}
	
	my $url = url $param, $this->{BASE};
	return $url->abs->as_string;
}


sub build_dom_tree
{
	my $this = shift;

	# XML::LibXML parser.
	$Cognition::HTMLParser::XDP = XML::LibXML->new();
	
	# Attempt to parse.
	if (!eval {
		$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
	})
	
	# If parsing failed, then 
	{
		$this->{_raw_source} = $this->{Source};
		
		# Figure out encoding.
		my $enc = 'iso-8859-1';
		if ($this->{RESPONSE}->header('content-type') =~ /\bcharset=[\"\'\s]?([a-z0-9-]+)/i
		|| substr($this->{Source},0,2048) =~ /\bencoding=[\"\'\s]?([a-z0-9-]+)/i 
		|| substr($this->{Source},0,2048) =~ /\bcharset=[\"\'\s]?([a-z0-9-]+)/i )
			{ $enc = lc($1); }
		
		# Decode into UTF-8.
		if ($enc ne 'utf-8')
		{
			$this->{decoded_from} = $enc;
			$this->{Source} = decode($enc, $this->{Source});
		}

		# Try parsing again.
		if (!eval {
			$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
		})
		
		# If it still fails, then
		{
			# Last ditch attempt at getting this into usable XHTML...
			$this->{Source} = $this->html2xhtml($this->{Source});
			
			# If we can't parse it yet, then die.
			if (!eval {
				$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
			})
			{
				print "COULD NOT PARSE!\n";
				die;
			}
		}
	}	
}


sub find_base_uri
{
	my $this = shift;

	# HTML <base> element.
	$this->{BASE}   = $this->{RESPONSE}->base();
	my @bases = $this->{DOM}->getElementsByTagName('base');
	my $base;
	foreach my $b (@bases)
	{
		$base = $b->getAttribute('href')
			if (length $b->getAttribute('href'));
	}
	$this->{BASE} = $this->abs_url($base)
		if (length $base);
}


sub find_namespaces
# Discover namespaces for metadata conforming to RFC 2731. Note
# that eRDF uses the same format for metadata schemas, so we
# can re-use these for eRDF too.
{
	my $this = shift;
	
	# Pull out XML namespaces defined on the root node. For RDFa, etc we'll
	# also read namespaces from other nodes, but not at this point.
	foreach my $a ($this->{DOM}->getDocumentElement->getAttributes)
	{
		my $k = $a->getName;
		my $v = $a->getValue;
		next unless $k =~ /^xmlns\:/i;
		$k =~ s/^xmlns\://i;
	
		$this->mergeNS($k, $v, "The '$k' Namespace", '', 0);
	}
	
	# Find custom RFC 2731 namespaces. FOAF and other stuff goes here.
	foreach my $l ($this->{DOM}->getElementsByTagName('link'))
	{
		my $rel = $l->getAttribute('rel');
		
		if ($rel =~ /^schema[\.\:]([A-Z_][A-Z0-9_-]*)$/i)
		{
			my $prefix = $1;
			my $url    = $this->abs_url($l->getAttribute('href'));
			my $title  = $l->getAttribute('title');
			
			$this->mergeNS($prefix, $url, ($title||"The '$prefix' Schema"), undef, 1);
		}
	}

	# Well-known namespaces, some defined so that people can be lazy. Others
	# defined because they're not usually used in an RDF-context, but I thought
	# it might be nice to try.
	$this->mergeNS('dcterms', 'http://purl.org/dc/terms/', 'Dublin Core Terms', -1, 1);
	$this->mergeNS('dcmitype', 'http://purl.org/dc/dcmitype/', 'Dublin Core Types', -1, 1);
	$this->mergeNS('dc', 'http://purl.org/dc/elements/1.1/', 'Dublin Core 1.1', -1, 1);
	$this->mergeNS('foaf', 'http://xmlns.com/foaf/0.1/', 'Friend of a Friend', -1, 1);
	$this->mergeNS('http11', 'urn:ietf:rfc:2616#', 'Hypertext Transfer Protocol 1.1', -1, 1);
	$this->mergeNS('iso8601', 'urn:iso:std:iso:8601#', 'ISO 8601', -1, 1);
	$this->mergeNS('openid', 'http://openid.net/specs/openid-authentication-1_1.html#', 'OpenID 1.1', -1, 1);
	$this->mergeNS('owl', 'http://www.w3.org/2002/07/owl#', 'OWL Web Ontology Language', -1, 1);
	$this->mergeNS('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'RDF/XML Syntax Specification (Revised)', -1, 1);
	$this->mergeNS('rdfs', 'http://www.w3.org/2000/01/rdf-schema#', 'RDF Vocabulary Description Language 1.0', -1, 1);
	$this->mergeNS('schema', 'urn:ietf:rfc:2731#', "The 'Schema' Metaschema", -1, 1);
	$this->mergeNS('service', 'http://www.w3.org/2007/app#', 'Atom Publishing Protocol', -1, 1);
	$this->mergeNS('link', 'http://www.w3.org/2006/link#', 'Link Vocab', -1, 1);
	$this->mergeNS('xhtml', 'http://www.w3.org/1999/xhtml#', 'XHTML', -1, 1);
	$this->mergeNS('xhv', 'http://www.w3.org/1999/xhtml/vocab#', 'XHTML Link Vocabulary', -1, 1);
	$this->mergeNS('xsd', 'http://www.w3.org/2001/XMLSchema#', 'XML Schema', -1, 1);
	
}


sub find_page_profiles
{
	my $this = shift;
	
	foreach my $head ($this->{DOM}->getElementsByTagName('head'))
	{
		my $profile = $head->getAttribute('profile');
		$profile =~ s/(^\s+|\s+$)//g;
		foreach my $url (split /\s+/s, $profile)
		{
			$url = $this->abs_url($url);
			$this->{Profiles}->{ $url } = 1;
		}
	}
	
	foreach my $link (Cognition::HTMLParser::Utils::searchRel('profile', $this->{DOM}))
	{
		my $url = $this->abs_url($link->getAttribute('href'));
		$this->{Profiles}->{ $url } = 1;
	}	
}


sub html2xhtml
{
	my $this = shift;
	my $html = shift;
	
	$html =~ s/\<\!DOCTYPE[^>]*\>//i;
	
	# Additional valid tags.
	$HTML::Tagset::isPhraseMarkup{'time'} = 1;
	$HTML::Tagset::is_Possible_Strict_P_Content{'time'} = 1;
	$HTML::Tagset::isBodyElement{'time'} = 1;
	$HTML::Tagset::isKnown{'time'} = 1;
	
	my $t = HTML::TreeBuilder->new();
	$t->implicit_tags(1);
	$t->p_strict(1);
	$t->parse_content($html);
	$t->{'xmlns'} = 'http://www.w3.org/1999/xhtml';
	
	my $xhtml = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
		. "<!DOCTYPE html\n"
		. "\tPUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
		. "\t\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
		. $t->as_XML(undef, "\t", {
			'area'=>1,
			'base'=>1,
			'basefont'=>1,
			'br'=>1,
			'col'=>1,
			'frame'=>1,
			'hr'=>1,
			'img'=>1,
			'input'=>1,
			'isindex'=>1,
			'link'=>1,
			'meta'=>1,
			'param'=>1
		});

	# XHTML now has some annoying named entities.
	#my %ents = %HTML::Entities::entity2char;
	#foreach my $e (keys %ents)
	#{
	#	my $char = $ents{$e};
	#	$e =~ s/\;$//;
	#	$xhtml =~ s/ \& $e \; / sprintf "&#x%X;", ord($char) /gex;
	#}
	
	return encode_entities($xhtml, "\x80-\xFF");
	
	return $xhtml;
}


sub load_stored_data
{
	my $this = shift;

	while (my $fn = shift)
	{
		my $file = $ENV{'HOME'}.'/.cognition/'.$fn.'.rdf';
		next unless (-x $file);
		
		open FILE, $file;
		my @lines = <FILE>;
		close FILE;
		my $rdf = join '', @lines;
		$this->parse_rdf_data($rdf, 'http://purl.org/dc/terms/', undef, 1);
	}

}


sub mergeNS
# Can be passed either an object with a metadata_ns method or
# a list of suggestedPrefix, url, title.
{
	my $this    = shift;
	my $mergeme = shift;
	
	if (ref $mergeme && UNIVERSAL::can($mergeme, 'metadata_ns'))
	{
		my $mns = $mergeme->metadata_ns;
		foreach my $k (keys %$mns)
		{
			$this->mergeNS($k, $mns->{$k}->{nsurl}, $mns->{$k}->{title}, undef, 0);
		}
	}
	
	else
	{
		my $suggestedPrefix = $mergeme;
		my $NS = Cognition::Namespace->new($suggestedPrefix, @_);
		
		my $addIt = 1;
		foreach my $existing (@{$this->{Namespaces}})
		{
			if (Cognition::Namespace::equiv($existing, $NS))
			{
				$addIt = 0;
				last;
			}
		}
		
		push @{$this->{Namespaces}}, $NS
			if ($addIt);

		return $suggestedPrefix;
	}
}


sub mergeSubject
{
	my $this    = shift;
	my $mergeme = shift;
	my $alt     = shift;
	my $target;

	my $key = $alt ? 'RDF1' : 'RDF';
	
	# Refuse to add boring subjects.
	return unless (defined $mergeme);
	return unless (((keys %{$mergeme->{properties}}) > 1)
	       ||      (1 < scalar(@{ $mergeme->{uris} })));
	
	foreach my $ex (keys %{$this->{$key}})
	{
		my $existing = $this->{$key}->{$ex};
		next unless (defined $existing && (ref $existing eq 'Cognition::RDFModel::Subject'));
		
		if ($existing->equal_to($mergeme))
		{
			$target = $ex;
			last;
		}
	}
	
	if (defined $target)
	{
		$this->{$key}->{$target}->merge($mergeme);
		$this->{$key}->{$target}->{page} = $this;
	}
	else
	{
		$this->{$key}->{ $mergeme->uri } = $mergeme;
		$this->{$key}->{ $mergeme->uri }->{page} = $this;
	}
}


sub parse_metadata
{
	my $this = shift;
	
	no strict 'refs';
	foreach my $module (qw(Metadata RoleAttr RDFa eRDF RDF::Embedded
		RDF::Linked RDF::Comment DocModel RDF::GRDDL))
	{
		my $is_enabled = 'Cognition::HTMLParser::'.$module.'::is_enabled';
		my $consume    = 'Cognition::HTMLParser::'.$module.'::consume';
		&{$consume}($this) if &{$is_enabled}($this);
	}
	use strict 'refs';

}


sub parse_uf
{
	my $this = shift;
	return unless ($this->{options}->{p_uf});
	
	no strict 'refs';
	foreach my $module (qw(hCard adr geo XFN RelTag RelLicense RelEnclosure
		hCalendar hAtom COinS species hMeasure hReview hResume figure hAudio))
	{
		my $consume    = 'Cognition::uF::'.$module.'::consume';
		&{$consume}($this);
	}
	use strict 'refs';
	
}

1;



######################################################################
package Cognition::Namespace;
######################################################################

sub new
{
	my $class = shift;
	my ($prefix, $uri, $title, $scope, $loose) = @_;
	my $this;
	
	$this = {
		prefix  => $prefix,
		uri     => $uri,
		title   => $title,
		scope   => $scope,
		loose   => $loose
	};
	
	bless $this, $class;
}


sub bestMatch
{
	my $prefix    = shift;
	my $xpath     = shift;	
	my $bestMatch = undef;
	my $highScore = -10;
	
	foreach my $ns (@_)
	{
		my $nsScore = -11;
		if ( ($ns->{scope}<0) || (!defined $ns->{scope}) )
		{
			if ($ns->{loose} && (lc($ns->{prefix}) eq lc($prefix)))
			{
				$nsScore = (defined $ns->{scope}) ? 0 : -2;
			}
			elsif ($ns->{prefix} eq $prefix)
			{
				$nsScore = (defined $ns->{scope}) ? 1 : -1;
			}
		}
		elsif ( (length $xpath)
		&& (substr($xpath,0,(length $ns->{scope})) eq $ns->{scope}) )
		{
			if ($ns->{loose} && (lc($ns->{prefix}) eq lc($prefix)))
			{
				$nsScore = length $ns->{scope};
			}
			elsif ($ns->{prefix} eq $prefix)
			{
				$nsScore = length $ns->{scope};
			}
		}
		
		if ($nsScore > $highScore)
		{
			$highScore = $nsScore;
			$bestMatch = $ns;
		}
	}
	
	return $bestMatch;
}


sub equiv
{
	my $existing = shift;
	my $new      = shift;
	
	if ($existing->{loose})
	{
		return 0 if (lc($existing->{prefix}) ne lc($new->{prefix}));
	}
	else
	{
		return 0 if ($existing->{prefix} ne $new->{prefix});
	}
	
	return 0 if ($existing->{uri} ne $new->{uri});
	return 1 if (!defined $existing->{scope});
	return 1 if ($existing->{scope} == -1);
	
	return 1
		if (((length $existing->{scope}) < (length $new->{scope}))
		&&  substr($new->{scope}, 0, (length $existing->{scope})) eq $existing->{scope});
		
	return 0;
}


1;
