#!/usr/bin/perl

######################################################################
package Cognition::HTMLParser::RDF;
######################################################################

use Cognition::MagicString;
use HTTP::Request;
use XML::LibXML qw(:all);
use XML::LibXSLT;

sub handle
{
	my $this = shift;
	my $rdf  = shift;
	my $base = shift || $this->{RESPONSE}->base;
	my $mime = shift || 'application/rdf+xml';
	my $alt  = shift;
	
	my $baseobj = new RDF::Redland::URI("$base");
	my $parser  = new RDF::Redland::Parser(undef, $mime);
	my $stream;
	
	if (!eval{$stream = $parser->parse_string_as_stream($rdf, $baseobj);})
	{
		# Funny little way of outputting an error message.
		my $S = Cognition::RDFModel::Subject->new('file:///dev/stderr');
		my $V = Cognition::RDFModel::Value->new("RDF parsing failed for:\n$rdf");
		$S->addVal('http://purl.org/dc/terms/abstract', $V);
		$this->mergeSubject($S);
		return;
	}

	while ($stream && !$stream->end)	
	{
		my $statement = $stream->current;
		
		my $subj = $statement->subject;
		my $prop = $statement->predicate;
		my $obj  = $statement->object;
		
		if ($subj->is_resource && $prop->is_resource)
		{
			my $S = Cognition::RDFModel::Subject->new($subj->uri->as_string);
			my $P = $prop->uri->as_string;
			my $V;
			
			if ($obj->is_resource)
				{ $V = Cognition::RDFModel::Value->new($obj->uri->as_string, 1); }
			else
				{ $V = Cognition::RDFModel::Value->new($obj->literal_value); }

			$S->addVal($P, $V);				
			$this->mergeSubject($S, $alt);	
		}
		
		$stream->next;
	}
	
	if ($RDF::Redland::VERSION gt '1.00_06_00')
	{
		my %ns = $parser->namespaces_seen;
		foreach my $k (keys %ns)
		{
			$this->mergeNS($k, $ns{$k}->as_string, "RDF $k", -1, 0);
		}
	}
	else
	{
		while ($rdf =~ /xmlns\:?([A-Za-z0-9_-]*)\=\"([^"]*)\"/g)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'defaultpfx';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix", -1, 0);
		}
		while ($rdf =~ /xmlns\:?([A-Za-z0-9_-]*)\=\'([^']*)\'/g)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'defaultpfx';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix", -1, 0);
		}
	}
} #/sub handle_rdf

1;


######################################################################
package Cognition::HTMLParser::RDF::Linked;
######################################################################

sub is_enabled
{
	my $this = shift;
	return $this->{options}->{p_rdfx};
}

sub consume
{
	my $this = shift;
	return unless ($this->{RDF}->{ $this->abs_url }
		->property('http://www.w3.org/1999/xhtml/vocab#meta'));
		
	foreach my $val ( @{ $this->{RDF}->{ $this->abs_url }
		->property('http://www.w3.org/1999/xhtml/vocab#meta')->valObjects } )
	{
		my $uri = URI->new($val->toString);
		next if ($this->{options}->{nofollow}->{ $uri->host });
	
		my $request  = HTTP::Request->new(GET => $val->toString);
		my $response = $this->{UA}->request($request);

		if ($response->header('content-type') =~ /(rdf)/i)
		{
			Cognition::HTMLParser::RDF::handle($this,
				$response->content,
				$response->base,
				$response->header('content-type'));
		}
	}
} #/sub consume

1;


######################################################################
package Cognition::HTMLParser::RDF::Comment;
######################################################################

sub is_enabled
{
	my $this = shift;
	return $this->{options}->{p_comments};
}

sub consume
{
	my $this = shift;
	my $src = $this->{_raw_source} || $this->{Source};

	while ($src =~ /<!--\s*(.*)\s*-->/sg)
	{
		my $comment = $1;
		Cognition::HTMLParser::RDF::handle($this, $comment)
			if ($comment =~ /^[\s\r\n]*<rdf:RDF/);
	}
}

1;


######################################################################
package Cognition::HTMLParser::RDF::Embedded;
######################################################################

sub is_enabled
{
	my $this = shift;
	return $this->{options}->{p_rdf};
}

sub consume
{
	my $this = shift;
	my @rdfs = $this->{DOM}->getElementsByTagNameNS("http://www.w3.org/1999/02/22-rdf-syntax-ns#", 'RDF');
	foreach my $rdf (@rdfs)
		{ Cognition::HTMLParser::RDF::handle($this, $rdf->toString); }
} #/sub consume

1;


######################################################################
package Cognition::HTMLParser::RDF::GRDDL;
######################################################################

sub is_enabled
{
	return 1;
}

sub consume
{
	my $this = shift;
	my @transformations;

	# Get a list of locally defined transformations.
	if ($this->{Profiles}->{'http://www.w3.org/2003/g/data-view'}
		|| !$this->{options}->{grddl_strict_profiles} )
	{
		my $prop = $this->{RDF}->{ $this->abs_url }
			->property('http://www.w3.org/2003/g/data-view#transformation');
		if ($prop)
		{
			foreach my $v (@{$prop->valObjects})
			{
				my $uri = URI->new($v->toString);
				next if ($this->{options}->{nofollow}->{ $uri->host });
				
				push @transformations, $v->toString if ($v->type);
			}
		}
	}
	
	# Get a list of transformations defined in the profiles.
	if ($this->{options}->{grddl_fetch})
	{
		foreach my $p (keys %{$this->{Profiles}})
		{
			my $uri = URI->new($p);
			next if ($this->{options}->{nofollow}->{ $uri->host });
		
			my $request  = HTTP::Request->new(GET => $p);
			my $response = $this->{UA}->request($request);

			if ($response->header('content-type') =~ /(html|xml)/i)
			{
				$this->{Profiles}->{$p} = Cognition::HTMLParser->new(
					$response->content,
					{
						ua        => $this->{UA},
						request   => $request,
						response  => $response,
						p_AUTO    => 0
					});
				my @trans   = Cognition::HTMLParser::Utils::searchRel(
					'profileTransformation', 
					$this->{Profiles}->{$p}->{DOM}
				);
				foreach my $t (@trans)
				{
					my $url = url $t->getAttribute('href'), $response->base;
					push @transformations, $url->abs->as_string
						unless ($this->{options}->{nofollow}->{ $url->host });
				}
			}
		}
	}
	
	my $attr = $this->_fq2pfx('http://www.w3.org/2003/g/data-view#transformation');
	my $attv = $this->{DOM}->documentElement->getAttribute($attr);
	$attv =~ s/(^\s+|\s+$)//g;
	$attv =~ s/\s+/ /g;
	my @attv = split / /, $attv;
	foreach my $a (@attv)
		{ push @transformations, $this->abs_url($a); }
	
	foreach my $t (@transformations)
	{
		my $uri = URI->new($t);
		next if ($this->{options}->{nofollow}->{ $uri->host });
	
		my $request  = HTTP::Request->new(GET => $t);
		my $response = $this->{UA}->request($request);

		if ($response->header('content-type') =~ /(xslt|xml)/i)
		{
			my $parser     = XML::LibXML->new();
			my $xslt       = XML::LibXSLT->new();
			my $style_doc  = $parser->parse_string($response->content);
			my $stylesheet = $xslt->parse_stylesheet($style_doc);
			my $results    = $stylesheet->transform($this->{DOM});
			my $rdfxml     = $stylesheet->output_string($results);
			
			Cognition::HTMLParser::RDF::handle($this, $rdfxml);
		}
	}
}

1;
